import { createSlice, PayloadAction } from '@reduxjs/toolkit';
import type { RootState } from './store';
import { GraphStatistics, retrieveGraphStatistics } from '../utils/result-statistics';
import { QueryMultiGraph } from '../../querybuilder';

export interface GraphQueryResultFromBackendPayload {
  queryID: string;
  result: {
    payload: GraphQueryResultFromBackend;
    type: string;
  };
}

export type NodeAttributes = { [key: string]: unknown };

export interface GraphQueryResultFromBackend {
  nodes: {
    id: string;
    label?: string;
    attributes: NodeAttributes;
  }[];

  edges: {
    id: string;
    label?: string;
    attributes: NodeAttributes;
    from: string;
    to: string;
  }[];

  // TODO: Also include type in node and edge
  // TODO: The backend should send all the different entitytypes and relationtypes in the result
}

export interface Node {
  id: string;
  label: string;
  attributes: NodeAttributes;
  mldata?: any; // FIXME
  /* type: string[]; */
}
export interface Edge {
  attributes: NodeAttributes;
  from: string;
  to: string;
  id: string;
  label: string;
  /* type: string; */
}

// Define a type for the slice state
export interface GraphQueryResult {
  nodes: Node[];
  edges: Edge[];

  // Describes what entities there are in this graph query result.
  nodeTypes: string[];
  queryingBackend: boolean;
  statistics?: GraphStatistics;
}

// Define the initial state using that type
export const initialState: GraphQueryResult = {
  nodes: [],
  edges: [],
  nodeTypes: [],
  queryingBackend: false,
};

export const graphQueryResultSlice = createSlice({
  name: 'graphQueryResult',
  // `createSlice` will infer the state type from the `initialState` argument
  initialState,
  reducers: {
    assignNewGraphQueryResult: (state, action: PayloadAction<GraphQueryResultFromBackendPayload>) => {
      const payload = action.payload.result.payload;

      // Only keep one node and one edge per id. This is also done in the backend, but we do it here as well to be sure.
      const nodeIDs = new Set(payload.nodes.map((node) => node.id));
      const edgeIDs = new Set(payload.edges.map((edge) => edge.id));
      let nodes = [...nodeIDs].map((nodeID) => payload.nodes.find((node) => node.id === nodeID) as unknown as Node);
      let edges = [...edgeIDs].map((edgeID) => payload.edges.find((edge) => edge.id === edgeID) as unknown as Edge);

      // Collect all the different nodetypes in the result
      const nodeTypes: string[] = [];
      nodes = nodes.map((node) => {
        let _node = { ...node };
        let nodeType = node.id.split('/')[0];
        let innerLabels = node?.attributes?.labels as string[];
        if (innerLabels?.length > 0) {
          nodeType = innerLabels[0] as string;
        }
        if (!nodeTypes.includes(nodeType)) nodeTypes.push(nodeType);
        _node.label = nodeType;
        return _node;
      });

      edges = edges.map((edge) => {
        let _edge = { ...edge };
        let edgeType = edge.id.split('/')[0];
        if (!_edge.id.includes('/')) {
          edgeType = edge.attributes.Type as string;
        }
        _edge.label = edgeType;
        return _edge;
      });

      // Assign new state
      state.nodes = nodes;
      state.edges = edges;
      state.nodeTypes = nodeTypes;
      state.queryingBackend = false;
    },
    resetGraphQueryResults: (state) => {
      // Assign new state
      state.nodes = [];
      state.edges = [];
      state.nodeTypes = [];
      state.queryingBackend = false;
    },
    queryingBackend: (state) => {
      state.queryingBackend = true;
    },
    addGraphQueryStatistics: (state, action: PayloadAction<QueryMultiGraph>) => {
      if (state.nodes.length > 0 && state.edges.length > 0) {
        state.statistics = retrieveGraphStatistics({ nodes: state.nodes, edges: state.edges }, action.payload);
      }
    },
  },
});

export const { assignNewGraphQueryResult, resetGraphQueryResults, queryingBackend, addGraphQueryStatistics } =
  graphQueryResultSlice.actions;

// Other code such as selectors can use the imported `RootState` type
export const selectGraphQueryResult = (state: RootState) => state.graphQueryResult;
export const selectGraphQueryResultNodes = (state: RootState) => state.graphQueryResult.nodes;
export const selectGraphQueryResultLinks = (state: RootState) => state.graphQueryResult.edges;
export const selectGraphQueryResultStats = (state: RootState) => state.graphQueryResult.statistics || {};

export default graphQueryResultSlice.reducer;
