import { createSelector, createSlice, PayloadAction } from '@reduxjs/toolkit';
import type { RootState } from './store';
import { globalConfigTypes, VisualizationConfiguration } from '../../vis/types';
import { isEqual } from 'lodash-es';
import { EncodingTypes } from '../../vis/configuration/encodings';
import { SettingTypes } from '../../vis/configuration/settings';
import { InteractionTypes } from '../../vis/configuration/interactions';

export type VisState = {
  activeVisualization: string;
  settings: {
    general: globalConfigTypes;
  } & {
    [id: string]: VisualizationConfiguration;
  };
};

export const initialState: VisState = {
  activeVisualization: 'NodeLink',
  settings: {
    general: {},
  },
};

export const visualizationSlice = createSlice({
  name: 'visualization',
  initialState,
  reducers: {
    addVisualization: (
      state,
      action: PayloadAction<{
        id: string;
        settings: SettingTypes;
        encodings: EncodingTypes;
        interactions: InteractionTypes;
      }>,
    ) => {
      const { id, settings, encodings, interactions } = action.payload;
      state.settings[id] = { settings, encodings, interactions };
    },
    removeVisualization: (state, action: PayloadAction<string>) => {
      if (state.settings[action.payload]) {
        delete state.settings[action.payload];
      }
    },
    updateVisualizationSettings: (state, action: PayloadAction<{ id: string; configuration: any }>) => {
      const { id, configuration } = action.payload;
      state.settings[id] = { ...state.settings[id], ...configuration };
    },
    updateGeneralSettings: (state, action: PayloadAction<globalConfigTypes>) => {
      state.settings.general = action.payload;
    },
    setActiveVisualization: (state, action: PayloadAction<string>) => {
      state.activeVisualization = action.payload;
    },
    setVisualizationState: (state, action: PayloadAction<VisState>) => {
      if (action.payload.activeVisualization && !isEqual(action.payload, state)) {
        state.activeVisualization = action.payload.activeVisualization;
        state.settings = action.payload.settings;
      }
    },
    updateGeneralConfiguration: (state, action: PayloadAction<any>) => {
      state.settings.general = action.payload;
    },
    updateVisConfiguration: (state, action: PayloadAction<any>) => {
      state.settings[state.activeVisualization] = action.payload;
    },
    updateVisSettings: (state, action: PayloadAction<any>) => {
      if (state.activeVisualization && state.settings[state.activeVisualization]) {
        state.settings[state.activeVisualization]!.settings = action.payload;
      }
    },
    updateVisEncodings: (state, action: PayloadAction<any>) => {
      if (state.activeVisualization && state.settings[state.activeVisualization]) {
        state.settings[state.activeVisualization]!.encodings = action.payload;
      }
    },
    updateVisInteractions: (state, action: PayloadAction<any>) => {
      if (state.activeVisualization && state.settings[state.activeVisualization]) {
        state.settings[state.activeVisualization]!.interactions = action.payload;
      }
    },
  },
});

export const {
  addVisualization,
  removeVisualization,
  updateVisualizationSettings,
  updateGeneralSettings,
  setActiveVisualization,
  setVisualizationState,
  updateGeneralConfiguration,
  updateVisConfiguration,
  updateVisSettings,
  updateVisEncodings,
  updateVisInteractions,
} = visualizationSlice.actions;

export const visualizationState = (state: RootState) => state.visualize;
export const visualizationAllSettings = (state: RootState) => state.visualize.settings;
export const visualizationSettings = createSelector([visualizationAllSettings, visualizationState], (settings, visualize) => ({
  general: settings.general,
  [visualize.activeVisualization]: visualize.settings[visualize.activeVisualization],
}));

export const visGeneralSettings = (state: RootState) => state.visualize.settings.general ?? {};
export const visSettingsState = (state: RootState) => state.visualize.settings[state.visualize.activeVisualization]?.settings ?? {};
export const visEncodingsState = (state: RootState) => state.visualize.settings[state.visualize.activeVisualization]?.encodings ?? {};
export const visInteractionsState = (state: RootState) => state.visualize.settings[state.visualize.activeVisualization]?.interactions ?? {};

export default visualizationSlice.reducer;
