import { useAppDispatch, useGraphQueryResult, useSchemaGraph } from '@graphpolaris/shared/lib/data-access/store';
import { useEffect, useRef, useState } from 'react';
import { AxisLabel, EntitiesFromSchema, MinMaxType, PlotSpecifications, PlotType, RelationType } from './Types';
import { isNodeLinkResult } from '../../shared/ResultNodeLinkParserUseCase';
import styles from './semanticsubstrates.module.scss';
import AddPlotButtonComponent from './subcomponents/AddPlotButtonComponent';
import SVGCheckboxesWithSemanticSubstrLabel from './subcomponents/SVGCheckBoxComponent';
import AddPlotPopup from './subcomponents/AddPlotPopup';
import VisConfigPanelComponent from '../../shared/VisConfigPanel/VisConfigPanel';
import Plot from './subcomponents/PlotComponent';
import LinesBetweenPlots from './subcomponents/LinesBetweenPlotsComponent';
import Color from 'color';
import CalcEntityAttrNamesFromResultUseCase from './utils/CalcEntityAttrNamesFromResultUseCase';
import CalcEntityAttrNamesFromSchemaUseCase from './utils/CalcEntityAttrNamesFromSchemaUseCase';
import CalcDefaultPlotSpecsUseCase from './utils/CalcDefaultPlotSpecsUseCase';
import ToPlotDataParserUseCase from './utils/ToPlotDataParserUseCase';
import CalcXYMinMaxUseCase from './utils/CalcXYMinMaxUseCase';
import CalcScaledPosUseCase from './utils/CalcScaledPositionsUseCase';
import { useImmer } from 'use-immer';
import FilterUseCase from './utils/FilterUseCase';

export type SemanticSubstrateState = {
  plotSpecifications: PlotSpecifications[];

  nodeRadius: number;
  nodeColors: string[];

  gapBetweenPlots: number;

  entitiesFromSchemaPruned: EntitiesFromSchema;

  selectedAttributeNumerical: string;
  selectedAttributeCatecorigal: string;
  relationSelectedAttributeNumerical: string;
  relationSelectedAttributeCatecorigal: string;
  scaleCalculation: (x: number) => number;
  colourCalculation: (x: string) => string;

  relationScaleCalculation: (x: number) => number;
  relationColourCalculation: (x: string) => string;

  addPlotButtonAnchor: (EventTarget & SVGRectElement) | null;
};

export type SemanticSubstratePlotState = {
  plots: PlotType[];

  // Relations is a 3d array, with the first array denoting the outbound plot, the second array denotes the inbound plot
  // For example, [plot1][plot2] will give all edges which are outbound from plot1 and ingoing to plot2
  allRelations: RelationType[][][];
  filteredRelations: RelationType[][][];

  // Determines if connections [fromPlotIndex][toPlotIndex] will be shown
  visibleRelations: boolean[][];

  // Used for filtering the relations when brushing (brush is the square selection tool)
  filtersPerPlot: { x: MinMaxType; y: MinMaxType }[];
};

/**
 * These functions are mock function for now, but can be properly implemented later down the line.
 */
const scaleCalculation: (x: number) => number = (x: number) => {
  return 3;
};
const colourCalculation: (x: string) => string = (x: string) => {
  return '#d56a50';
};
const relationColourCalculation: (x: string) => string = (x: string) => {
  return '#d49350';
};
const relationScaleCalculation: (x: number) => number = (x: number) => {
  return 1;
};

export type SemanticSubstratesProps = {};

export const SemanticSubstrates = ({}: SemanticSubstratesProps) => {
  const graphQueryResult = useGraphQueryResult();
  const schema = useSchemaGraph();
  const [entitiesFromSchema, setEntitiesFromSchema] = useState<EntitiesFromSchema>({ entityNames: [], attributesPerEntity: {} });

  const [state, setState] = useImmer<SemanticSubstrateState>({
    plotSpecifications: [],
    entitiesFromSchemaPruned: { entityNames: [], attributesPerEntity: {} },
    selectedAttributeNumerical: 'long',
    selectedAttributeCatecorigal: 'state',
    relationSelectedAttributeNumerical: 'Distance',
    relationSelectedAttributeCatecorigal: 'Day',
    nodeRadius: 3,
    nodeColors: ['#D56A50', '#1E9797', '#d49350', '#1e974a', '#D49350'],
    gapBetweenPlots: 50,
    scaleCalculation: scaleCalculation,
    colourCalculation: colourCalculation,
    relationScaleCalculation: relationScaleCalculation,
    relationColourCalculation: relationColourCalculation,

    addPlotButtonAnchor: null,
  });

  const [plotState, setPlotState] = useImmer<SemanticSubstratePlotState>({
    allRelations: [],
    filteredRelations: [],
    plots: [],
    filtersPerPlot: [],
    visibleRelations: [],
  });

  useEffect(() => {
    const ret = CalcEntityAttrNamesFromSchemaUseCase.calculate(schema);
    setEntitiesFromSchema(ret);
  }, [schema]);

  useEffect(() => {
    if (isNodeLinkResult(graphQueryResult)) {
      // Only apply new result if we have a valid schema
      if (entitiesFromSchema.entityNames.length === 0) {
        console.log('Semantic substrates: No valid schema available.');
        return;
      }

      setState((draft) => {
        draft.entitiesFromSchemaPruned = CalcEntityAttrNamesFromResultUseCase.CalcEntityAttrNamesFromResult(
          graphQueryResult,
          entitiesFromSchema
        );

        // Generate default plots, if there aren't any currently
        if (state.plotSpecifications.length == 0) {
          draft.plotSpecifications = CalcDefaultPlotSpecsUseCase.calculate(graphQueryResult);
        }

        return draft;
      });
    } else {
      console.error('Invalid query result!');
    }

    applyNewPlotSpecifications();
  }, [graphQueryResult, entitiesFromSchema]);

  useEffect(() => {
    applyNewPlotSpecifications();
  }, [state]);

  /**
   * Update the scaling of the nodes after a new attribute has been selected. This function is called when the new attribute has numbers as values.
   * @param scale The scaling function provided by the config panel ViewModel.
   */
  function changeSelectedAttributeNumerical(scale: (x: number) => number): void {
    setState({ ...state, scaleCalculation: scale });
  }

  /**
   * Update the colours of the nodes after a new attribute has been selected. This function is called when the new attribute has NaNs as values.
   * @param getColour The colouring function provided by the config panel ViewModel.
   */
  function changeSelectedAttributeCatecorigal(getColour: (x: string) => string): void {
    setState({ ...state, colourCalculation: getColour });
  }

  /**
   * Update the scaling of the nodes after a new relation attribute has been selected. This function is called when the new relation attribute has numbers as values.
   * @param scale The scaling function provided by the config panel ViewModel.
   */
  function changeRelationSelectedAttributeNumerical(scale: (x: number) => number): void {
    setState({ ...state, relationScaleCalculation: scale });
  }

  /**
   * Update the colours of the nodes after a new relation attribute has been selected. This function is called when the new relation attribute has NaNs as values.
   * @param getColour The colouring function provided by the config panel ViewModel.
   */
  function changeRelationSelectedAttributeCatecorigal(getColour: (x: string) => string): void {
    setState({ ...state, relationColourCalculation: getColour });
  }

  /**
   * Apply plot specifications to a node link query result. Calculates the plots and relations using the provided plot specs.
   * @param {PlotSpecifications[]} plotSpecs The plotspecs to filter the new plots with.
   * @param {NodeLinkResultType} queryResult The query result to apply the plot specs to.
   */
  function applyNewPlotSpecifications(): void {
    // Parse the incoming data to plotdata with the auto generated plot specifications
    const { plots, relations } = ToPlotDataParserUseCase.parseQueryResult(
      graphQueryResult,
      state.plotSpecifications,
      state.relationSelectedAttributeNumerical,
      state.relationSelectedAttributeCatecorigal,
      state.relationScaleCalculation,
      state.relationColourCalculation
    );

    setPlotState((draft) => {
      draft.allRelations = relations;
      // Clone relations to filteredRelations, because the filters are initialized with no constraints
      draft.filteredRelations = relations.map((i) => i.map((j) => j.map((v) => v)));

      // Calculate the scaled positions to the width and height of a plot
      let yOffset = 0;
      draft.plots = plots.map((plot, i) => {
        const minmaxAxis = CalcXYMinMaxUseCase.calculate(plot);
        const scaledPositions = CalcScaledPosUseCase.calculate(plot, minmaxAxis.x, minmaxAxis.y);

        // Collect all possible values for a certain attribute, used for the autocomplete title field
        const possibleTitleAttributeValues = graphQueryResult.nodes.reduce((values: Set<string>, node) => {
          // Filter on nodes which are of the entity type in the plotspec
          // Use a Set so we only collect unique values
          if (node.label == state.plotSpecifications[i].entity && state.plotSpecifications[i].labelAttributeType in node.attributes)
            values.add(node.attributes[state.plotSpecifications[i].labelAttributeType] as string);
          return values;
        }, new Set<string>());

        // Accumulate the yOffset for each plot, with the width and gapbetweenplots
        const thisYOffset = yOffset;
        yOffset += plot.height + state.gapBetweenPlots;
        return {
          title: plot.title,
          nodes: plot.nodes.map((node, i) => {
            return { ...node, scaledPosition: scaledPositions[i] };
          }),
          selectedAttributeNumerical: state.selectedAttributeNumerical,
          scaleCalculation: state.scaleCalculation,
          selectedAttributeCatecorigal: state.selectedAttributeCatecorigal,
          colourCalculation: state.colourCalculation,
          minmaxXAxis: minmaxAxis.x,
          minmaxYAxis: minmaxAxis.y,
          width: plot.width,
          height: plot.height,
          yOffset: thisYOffset,
          possibleTitleAttributeValues: Array.from(possibleTitleAttributeValues),
        };
      });

      // Initialize filters with no constraints
      draft.filtersPerPlot = plotState.plots.map((plot) => {
        return {
          x: { min: plot.minmaxXAxis.min, max: plot.minmaxXAxis.max },
          y: { min: plot.minmaxYAxis.min, max: plot.minmaxYAxis.max },
        };
      });

      // Initialize the visible relations with all false values
      if (draft.visibleRelations.length !== draft.plots.length) {
        draft.visibleRelations = new Array(draft.plots.length).fill([]).map(() => new Array(draft.plots.length).fill(false));
      }

      return draft;
    });
  }

  /**
   * Updates the visualisation when a checkbox is pressed.
   * @param fromPlotIndex Which plot the connections should come from.
   * @param toPlotIndex Which plot the connections should go to.
   * @param value Whether the box has been checked or not.
   */
  function onCheckboxChanged(fromPlotIndex: number, toPlotIndex: number, value: boolean): void {
    setPlotState((draft) => {
      draft.visibleRelations[fromPlotIndex][toPlotIndex] = value;
      return draft;
    });
  }

  /** Callback for the square selection tool on the plots.
   * Changes the filter of the plot index and applies this new filter.
   * @param {number} plotIndex The index of the plot where there is brushed.
   * @param {MinMaxType} xRange The min and max x values of the selection.
   * @param {MinMaxType} yRange The min and max y values of the selection.
   */
  function onBrush(plotIndex: number, xRange: MinMaxType, yRange: MinMaxType) {
    setPlotState((draft) => {
      // Change the filter of the corresponding plot
      draft.filtersPerPlot[plotIndex] = {
        x: xRange,
        y: yRange,
      };

      // Apply the new filter to the relations
      FilterUseCase.filterRelations(draft.allRelations, draft.filteredRelations, draft.plots, draft.filtersPerPlot, plotIndex);
      return draft;
    });
  }

  /**
   * Prepends a new plot specifications and re-applies this to the current nodelink query result.
   * @param {string} entity The entity filter for the new plot.
   * @param {string} attributeName The attribute to filter on for the new plot.
   * @param {string} attributeValue The value for the attributeto filter on for the new plot.
   */
  function addPlot(entity: string, attributeName: string, attributeValue: string): void {
    setState((draft) => {
      draft.plotSpecifications = [
        {
          entity: entity,
          labelAttributeType: attributeName,
          labelAttributeValue: attributeValue,
          xAxis: AxisLabel.evenlySpaced, // Use default evenly spaced and # outbound connections on the x and y axis
          yAxis: AxisLabel.outboundConnections,
          xAxisAttributeType: '',
          yAxisAttributeType: '',
          width: 800, // The default width and height are 800x200
          height: 200,
        },
        ...state.plotSpecifications,
      ];

      return draft;
    });
  }

  /**
   * Callback when the delete button of a plot is pressed.
   * Will remove values at plotIndex from the `plotSpecifications`, `plots` , `filtersPerPlot` , `relations` , `filteredRelations` and `visibleRelations`.
   * @param {number} plotIndex The index for the plot that needs to be deleted.
   */
  function onDelete(plotIndex: number): void {
    setState((draft) => {
      draft.plotSpecifications.splice(plotIndex, 1);
      return draft;
    });

    setPlotState((draft) => {
      // Recalculate the plot y offsets with this plot removed.
      for (let i = plotIndex + 1; i < draft.plots.length; i++)
        draft.plots[i].yOffset -= draft.plots[plotIndex].height + state.gapBetweenPlots;

      draft.plots.splice(plotIndex, 1);
      draft.filtersPerPlot.splice(plotIndex, 1);

      draft.allRelations.splice(plotIndex, 1);
      draft.allRelations = draft.allRelations.map((r) => r.filter((_, i) => i != plotIndex));
      draft.filteredRelations.splice(plotIndex, 1);
      draft.filteredRelations = draft.filteredRelations.map((r) => r.filter((_, i) => i != plotIndex));
      draft.visibleRelations.splice(plotIndex, 1);
      draft.visibleRelations = draft.visibleRelations.map((r) => r.filter((_, i) => i != plotIndex));

      return draft;
    });
  }

  /**
   * Changes the axislabel of a plot.
   * @param {number} plotIndex The plot number for which to change an axis.
   * @param {'x' | 'y'} axis The axis to change. Can only be 'x' or 'y'.
   * @param {string} newLabel The axis label.
   */
  function onAxisLabelChanged(plotIndex: number, axis: 'x' | 'y', newLabel: string): void {
    setState((draft) => {
      const axisLabels: string[] = Object.values(AxisLabel);
      if (axisLabels.includes(newLabel) && newLabel != AxisLabel.byAttribute) {
        // If the new axis label is one of "# outbound conn", "# inbound conn" or "evenly spaced"
        if (axis === 'x') draft.plotSpecifications[plotIndex].xAxis = newLabel as AxisLabel;
        else if (axis === 'y') draft.plotSpecifications[plotIndex].yAxis = newLabel as AxisLabel;
      } else {
        // Else it is an attribute of the entity
        if (axis === 'x') {
          draft.plotSpecifications[plotIndex].xAxis = AxisLabel.byAttribute;
          draft.plotSpecifications[plotIndex].xAxisAttributeType = newLabel;
        } else if (axis === 'y') {
          draft.plotSpecifications[plotIndex].yAxis = AxisLabel.byAttribute;
          draft.plotSpecifications[plotIndex].yAxisAttributeType = newLabel;
        }
      }

      return draft;
    });
  }

  /**
   * Applies the new plot filter. Called when the user changes the plot title/filter.
   * @param {number} plotIndex The plotindex of which the title is changed.
   * @param {string} entity The new entity value for the plot filter, might be unchanged.
   * @param {string} attrName The new attribute name for the plot filter, might be unchanged.
   * @param {string} attrValue The new attribute value for the plot filter.
   */
  function onPlotTitleChanged(plotIndex: number, entity: string, attrName: string, attrValue: string): void {
    // If the entity or the attrName changed, auto select a default attrValue
    if (
      (entity != state.plotSpecifications[plotIndex].entity || attrName != state.plotSpecifications[plotIndex].labelAttributeType) &&
      attrValue == state.plotSpecifications[plotIndex].labelAttributeValue
    ) {
      const firstValidNode = graphQueryResult.nodes.find((node) => node.label == entity && attrName in node.attributes);
      if (firstValidNode != undefined) attrValue = firstValidNode.attributes[attrName] as string;
    }

    setState((draft) => {
      draft.plotSpecifications[plotIndex] = {
        ...draft.plotSpecifications[plotIndex],
        entity,
        labelAttributeType: attrName,
        labelAttributeValue: attrValue,
      };

      return draft;
    });
  }

  const plotElements: JSX.Element[] = [];
  for (let i = 0; i < plotState.plots.length; i++) {
    if (!state.plotSpecifications?.[i]) continue;

    plotElements.push(
      <Plot
        key={plotState.plots[i].title + i}
        plotData={plotState.plots[i]}
        plotSpecification={state.plotSpecifications[i]}
        entitiesFromSchema={entitiesFromSchema}
        nodeColor={state.nodeColors[i]}
        nodeRadius={state.nodeRadius}
        width={plotState.plots[i].width}
        height={plotState.plots[i].height}
        selectedAttributeNumerical={plotState.plots[i].selectedAttributeNumerical}
        selectedAttributeCatecorigal={plotState.plots[i].selectedAttributeCatecorigal}
        scaleCalculation={plotState.plots[i].scaleCalculation}
        colourCalculation={plotState.plots[i].colourCalculation}
        onBrush={(xRange: MinMaxType, yRange: MinMaxType) =>
          // Call the onBrush method with the index of this plot
          onBrush(i, xRange, yRange)
        }
        onDelete={() => onDelete(i)}
        onAxisLabelChanged={(axis: 'x' | 'y', value: string) => onAxisLabelChanged(i, axis, value)}
        onTitleChanged={(entity: string, attrName: string, attrValue: string) => onPlotTitleChanged(i, entity, attrName, attrValue)}
      />
    );
  }

  // Create the arrow-lines between the plots.
  const linesBetween: JSX.Element[] = [];
  for (let i = 0; i < plotState.filteredRelations.length; i++) {
    for (let j = 0; j < plotState.filteredRelations[i].length; j++) {
      let color = Color(state.nodeColors[i]);
      color = i == j ? color.lighten(0.3) : color.darken(0.3);
      if (plotState.visibleRelations?.[i]?.[j]) {
        // Only create the lines if the checkbox is ticked
        linesBetween.push(
          <LinesBetweenPlots
            key={plotState.plots[i].title + i + '-' + plotState.plots[j].title + j}
            fromPlot={plotState.plots[i]}
            toPlot={plotState.plots[j]}
            relations={plotState.filteredRelations[i][j]}
            color={color.hex()}
            nodeRadius={state.nodeRadius}
          />
        );
      }
    }
  }

  useEffect(() => {}, [state]);

  const heightOfAllPlots =
    plotState.plots.length > 0
      ? plotState.plots[plotState.plots.length - 1].yOffset + plotState.plots[plotState.plots.length - 1].height + 50
      : 0;
  return (
    <div className={styles.container}>
      <div style={{ width: '100%', height: '100%' }}>
        <svg style={{ width: '100%', height: heightOfAllPlots + 60 }}>
          <AddPlotButtonComponent
            x={750}
            onClick={(event: React.MouseEvent<SVGRectElement>) => setState({ ...state, addPlotButtonAnchor: event.currentTarget })}
          />
          <g transform={'translate(60,60)'}>
            <SVGCheckboxesWithSemanticSubstrLabel
              plots={plotState.plots}
              relations={plotState.allRelations}
              visibleRelations={plotState.visibleRelations}
              nodeColors={state.nodeColors}
              onCheckboxChanged={(fromPlot: number, toPlot: number, value: boolean) => onCheckboxChanged(fromPlot, toPlot, value)}
            />
            {plotElements}
            {linesBetween},
          </g>
        </svg>
      </div>
      <AddPlotPopup
        open={Boolean(state.addPlotButtonAnchor)}
        anchorEl={state.addPlotButtonAnchor}
        entitiesFromSchema={entitiesFromSchema}
        nodeLinkResultNodes={graphQueryResult.nodes}
        handleClose={() => setState({ ...state, addPlotButtonAnchor: null })}
        addPlot={(entity: string, attributeName: string, attributeValue: string) => addPlot(entity, attributeName, attributeValue)}
      />
      <VisConfigPanelComponent>
        {/* <FSSConfigPanel
          fssViewModel={semanticSubstratesViewModel.current}
          graph={state.nodeLinkQueryResult}
        // currentColours={currentColours}
        /> */}
      </VisConfigPanelComponent>
    </div>
  );
};

export default SemanticSubstrates;
